@@ -18,7 +18,7 @@
 #include "logsum.h"
 
 #define PACKAGE_NAME "nanopolish"
-#define PACKAGE_VERSION "0.4.0"
+#define PACKAGE_VERSION "0.5.0"
 #define PACKAGE_BUGREPORT "https://github.com/jts/nanopolish/issues"
 
 //
@@ -108,4 +108,20 @@ inline double add_logs(const double a, const double b)
 // split a string based on a delimiter
 std::vector<std::string> split(std::string in, char delimiter);
 
+// returns true if the provided string ends with the given suffix
+bool ends_with(const std::string& str, const std::string& suffix);
+
+// from: http://stackoverflow.com/questions/9330915/number-of-combinations-n-choose-r-in-c
+size_t nChoosek(size_t n, size_t k);
+
+// print a warning message to stderr a single time
+// this is only for debugging, please don't litter the code with them
+#define WARN_ONCE(x) static bool _warn_once = true; if(_warn_once) \
+                     fprintf(stderr, "WARNING: [%s]\n", (x)); _warn_once = false;
+
+template<class T>
+std::string array2str(const T& array)
+{
+    return std::string(array.begin(), array.end());
+}
 #endif
