@@ -27,6 +27,7 @@
 #include "nanopolish_anchor.h"
 #include "nanopolish_fast5_map.h"
 #include "nanopolish_hmm_input_sequence.h"
+#include "nanopolish_pore_model_set.h"
 #include "profiler.h"
 #include "progress.h"
 #include "stdaln.h"
@@ -57,7 +58,7 @@ SUBPROGRAM " Version " PACKAGE_VERSION "\n"
 "Copyright 2015 Ontario Institute for Cancer Research\n";
 
 static const char *CONSENSUS_USAGE_MESSAGE =
-"Usage: " PACKAGE_NAME " " SUBPROGRAM " [OPTIONS] --reads reads.fa --bam alignments.bam --genome genome.fa\n"
+"Usage: " PACKAGE_NAME " " SUBPROGRAM " [OPTIONS] -w contig:start-end --reads reads.fa --bam alignments.bam --genome genome.fa\n"
 "Compute a new consensus sequence for an assembly using a signal-level HMM\n"
 "\n"
 "  -v, --verbose                        display verbose output\n"
@@ -69,6 +70,7 @@ static const char *CONSENSUS_USAGE_MESSAGE =
 "  -g, --genome=FILE                    the genome we are computing a consensus for is in FILE\n"
 "  -o, --outfile=FILE                   write result to FILE [default: stdout]\n"
 "  -t, --threads=NUM                    use NUM threads (default: 1)\n"
+"      --models-fofn=FILE               read alternative k-mer models from FILE\n"
 "\nReport bugs to " PACKAGE_BUGREPORT "\n\n";
 
 namespace opt
@@ -79,13 +81,15 @@ namespace opt
     static std::string genome_file;
     static std::string output_file;
     static std::string window;
+    static std::string models_fofn;
+    static std::string alternative_model_type = DEFAULT_MODEL_TYPE;
     static int show_progress = 0;
     static int num_threads = 1;
 }
 
 static const char* shortopts = "r:b:g:t:w:o:v";
 
-enum { OPT_HELP = 1, OPT_VERSION, OPT_VCF, OPT_PROGRESS };
+enum { OPT_HELP = 1, OPT_VERSION, OPT_VCF, OPT_PROGRESS, OPT_MODELS_FOFN };
 
 static const struct option longopts[] = {
     { "verbose",     no_argument,       NULL, 'v' },
@@ -95,6 +99,7 @@ static const struct option longopts[] = {
     { "window",      required_argument, NULL, 'w' },
     { "outfile",     required_argument, NULL, 'o' },
     { "threads",     required_argument, NULL, 't' },
+    { "models-fofn", required_argument, NULL, OPT_MODELS_FOFN },
     { "progress",    no_argument,       NULL, OPT_PROGRESS },
     { "help",        no_argument,       NULL, OPT_HELP },
     { "version",     no_argument,       NULL, OPT_VERSION },
@@ -126,6 +131,7 @@ std::vector<HMMInputData> get_input_for_columns(HMMRealignmentInput& window,
         if(start_sa.rc != end_sa.rc)
             continue;
 
+
         HMMInputData data;
 
         uint32_t read_idx = rsi / 2;
@@ -151,11 +157,7 @@ std::vector<HMMInputData> get_input_for_columns(HMMRealignmentInput& window,
 // scoring functions without writing a bunch of code
 double score_sequence(const std::string& sequence, const HMMInputData& data)
 {
-    //return score_skip_merge(sequence, state);
-    //return score_khmm_model_postmerge(sequence, state);
-    //return khmm_score(sequence, state, AP_GLOBAL);
     return profile_hmm_score(sequence, data);
-    //return score_emission_dp(sequence, state);
 }
 
 
@@ -220,7 +222,6 @@ bool sortIndexedPathScoreDesc(const IndexedPathScore& a, const IndexedPathScore&
 void score_paths(PathConsVector& paths, const std::vector<HMMInputData>& input)
 {
     PROFILE_FUNC("score_paths")
-    double MIN_FIT = INFINITY;
     size_t CULL_RATE = 5;
     double CULL_MIN_SCORE = -30.0f;
     double CULL_MIN_IMPROVED_FRACTION = 0.2f;
@@ -254,7 +255,7 @@ void score_paths(PathConsVector& paths, const std::vector<HMMInputData>& input)
             fprintf(stderr, "Scoring %d\n", ri);
         }
 
-        const HMMInputData& data = input[ri];
+        //const HMMInputData& data = input[ri];
         std::vector<IndexedPathScore> result(paths.size());
 
         // Score all paths
@@ -306,6 +307,7 @@ void score_paths(PathConsVector& paths, const std::vector<HMMInputData>& input)
     std::stable_sort(paths.begin(), paths.end(), sortPathConsScoreDesc);
 
 #if DEBUG_PATH_SELECTION
+    double MIN_FIT = INFINITY;
     for(size_t pi = 0; pi < paths.size(); ++pi) {
 
         // Calculate the length of the matching prefix with the initial sequence
@@ -313,20 +315,16 @@ void score_paths(PathConsVector& paths, const std::vector<HMMInputData>& input)
 
         char initial = s == first ? 'I' : ' ';
 
-        printf("%zu\t%s\t%.1lf\t%zu %c %s", pi, paths[pi].path.c_str(), paths[pi].score, paths[pi].sum_rank, initial, paths[pi].mutdesc.c_str());
+        fprintf(stderr, "%zu\t%s\t%.1lf\t%zu %c %s", pi, paths[pi].path.c_str(), paths[pi].score, paths[pi].sum_rank, initial, paths[pi].mutdesc.c_str());
         // If this is the truth path or the best path, show the scores for all reads
         if(pi <= 1 || initial == 'I') {
             for(uint32_t ri = 0; ri < input.size(); ++ri) {
                 const HMMInputData& data = input[ri];
-                const KHMMParameters& parameters = data.read->parameters[data.strand];
-                if( fabs(parameters.fit_quality) > MIN_FIT)
-                    continue;
-
                 double curr = score_sequence(paths[pi].path, input[ri]);
-                printf("%.1lf,%.2lf ", parameters.fit_quality, curr);
+                fprintf(stderr, "%.2lf ", curr);
             }
         }
-        printf("\n");
+        fprintf(stderr, "\n");
     }
 #endif
 
@@ -340,7 +338,7 @@ void extend_paths(PathConsVector& paths, int maxk = 2)
 
     for(int k = 1; k <= maxk; ++k) {
 
-        for(int pi = 0; pi < paths.size(); ++pi) {
+        for(unsigned pi = 0; pi < paths.size(); ++pi) {
     
             std::string first(k, 'A');
             std::string extension = first;
@@ -539,8 +537,10 @@ void filter_outlier_data(std::vector<HMMInputData>& input, const std::string& se
         if(opt::verbose >= 1) {
             fprintf(stderr, "OUTLIER_FILTER %d %.2lf %.2lf %.2lf\n", ri, curr, n_events, lp_per_event);
         }
-
-        if(fabs(lp_per_event) < 3.5f) {
+        
+        // R9 thresholds
+        double threshold = model_stdv() ? 8.0f : 4.0f; // TODO: check
+        if(fabs(lp_per_event) < threshold) {
             out_rs.push_back(rs);
         }
     }
@@ -560,7 +560,7 @@ std::string join_sequences_at_kmer(const std::string& a, const std::string& b, c
     return a + b.substr(k);
 }
 
-void run_splice_segment(HMMRealignmentInput& window, uint32_t segment_id)
+void run_splice_segment(HMMRealignmentInput& window, uint32_t segment_id, const uint32_t k)
 {
     // The structure of the data looks like this:
 
@@ -593,11 +593,11 @@ void run_splice_segment(HMMRealignmentInput& window, uint32_t segment_id)
 
     // set up the input data for the HMM
     std::vector<HMMInputData> data = get_input_for_columns(window, start_column, end_column);
-    
-    // assume models for all the reads have the same k
-    assert(!data.empty());
-    const uint32_t k = data[0].read->pore_model[data[0].strand].k;
 
+    if(opt::verbose > 0) {
+        fprintf(stderr, "correcting segment %u with %zu reads\n", segment_id, data.size());
+    }
+    
     // The current consensus sequence
     std::string original = join_sequences_at_kmer(s_m_base, m_e_base, k);
     std::string base = original;
@@ -669,8 +669,12 @@ void train_segment(HMMRealignmentInput& window, uint32_t segment_id)
     // Set up the the input data for the HMM
     std::vector<HMMInputData> input = get_input_for_columns(window, start_column, end_column);
 
+    // no training can be performed if there are no reads for this segment
+    if(input.empty()) {
+        return;
+    }
+
     // assume models for all the reads have the same k
-    assert(!input.empty());
     const uint32_t k = input[0].read->pore_model[input[0].strand].k;
 
     std::string segment_sequence = join_sequences_at_kmer(s_m_base, m_e_base, k);
@@ -699,18 +703,32 @@ void train(HMMRealignmentInput& window)
 std::string call_consensus_for_window(const Fast5Map& name_map, const std::string& contig, int start_base, int end_base)
 {
     const int minor_segment_stride = 50;
-    HMMRealignmentInput window = build_input_for_region(opt::bam_file, opt::genome_file, name_map, contig, start_base, end_base, minor_segment_stride);
+    HMMRealignmentInput window = build_input_for_region(opt::bam_file,
+                                                        opt::genome_file,
+                                                        name_map,
+                                                        contig,
+                                                        start_base,
+                                                        end_base,
+                                                        minor_segment_stride,
+                                                        opt::alternative_model_type);
+    uint32_t num_segments = window.anchored_columns.size();
 
-    if(window.reads.empty()) {
+    // If there are not reads or not enough segments do not try to call a consensus sequence
+    if(window.reads.empty() || num_segments < 3) {
         // No data for this window, just return the original sequence as the consensus
         assert(!window.original_sequence.empty());
         return window.original_sequence;
     }
+    
+    if(opt::verbose > 0) {
+        fprintf(stderr, "correcting window %s:%d-%d with %zu reads\n", contig.c_str(), start_base, end_base, window.reads.size());
+    }
 
     //
     // Train the HMM
     //
-    train(window);
+    WARN_ONCE("Debug: using default transition parameters"); 
+    //train(window);
 
     // assume models for all the reads have the same k
     const uint32_t k = window.reads[0]->pore_model[T_IDX].k;
@@ -721,7 +739,6 @@ std::string call_consensus_for_window(const Fast5Map& name_map, const std::strin
     std::string reference = "";
     std::string consensus = "";
 
-    uint32_t num_segments = window.anchored_columns.size();
     uint32_t start_segment_id = 0;
 
     // Copy the base segments before they are updated
@@ -744,7 +761,7 @@ std::string call_consensus_for_window(const Fast5Map& name_map, const std::strin
         }
 
         // run the consensus algorithm for this segment
-        run_splice_segment(window, segment_id);
+        run_splice_segment(window, segment_id, k);
 
         // run_splice_segment updates the base_sequence of the current anchor, grab it and append
         std::string base = window.anchored_columns[segment_id].base_sequence;
@@ -786,6 +803,7 @@ void parse_consensus_options(int argc, char** argv)
             case '?': die = true; break;
             case 't': arg >> opt::num_threads; break;
             case 'v': opt::verbose++; break;
+            case OPT_MODELS_FOFN: arg >> opt::models_fofn; break;
             case OPT_PROGRESS: opt::show_progress = 1; break;
             case OPT_HELP:
                 std::cout << CONSENSUS_USAGE_MESSAGE;
@@ -823,6 +841,19 @@ void parse_consensus_options(int argc, char** argv)
         std::cerr << SUBPROGRAM ": a --bam file must be provided\n";
         die = true;
     }
+    
+    if(opt::models_fofn.empty()) {
+        std::cerr << SUBPROGRAM ": a --models file must be provided\n";
+        die = true;
+    } else {
+        // initialize the model set from the fofn
+        PoreModelSet::initialize(opt::models_fofn);
+    }
+
+    if(opt::window.empty()) {
+        std::cerr << SUBPROGRAM ": the -w (window) parameter must be provided\n";
+        die = true;
+    }
 
     if (die) 
     {
@@ -837,7 +868,7 @@ int consensus_main(int argc, char** argv)
     omp_set_num_threads(opt::num_threads);
 
     Fast5Map name_map(opt::reads_file);
-
+    
     // Parse the window string
     // Replace ":" and "-" with spaces to make it parseable with stringstream
     std::replace(opt::window.begin(), opt::window.end(), ':', ' ');
@@ -872,4 +903,5 @@ int consensus_main(int argc, char** argv)
     if(out_fp != stdout) {
         fclose(out_fp);
     }
+    return 0;
 }
