@@ -3,7 +3,7 @@
 // Written by Jared Simpson (jared.simpson@oicr.on.ca)
 //---------------------------------------------------------
 //
-// nanopolish_methyltrain -- train a methylation model
+// nanopolish_methyltest -- test CpG sites for methylation
 //
 #include <stdio.h>
 #include <stdlib.h>
@@ -30,6 +30,7 @@
 #include "nanopolish_anchor.h"
 #include "nanopolish_fast5_map.h"
 #include "nanopolish_methyltrain.h"
+#include "nanopolish_pore_model_set.h"
 #include "H5pubconf.h"
 #include "profiler.h"
 #include "progress.h"
@@ -85,7 +86,7 @@ SUBPROGRAM " Version " PACKAGE_VERSION "\n"
 
 static const char *METHYLTEST_USAGE_MESSAGE =
 "Usage: " PACKAGE_NAME " " SUBPROGRAM " [OPTIONS] --reads reads.fa --bam alignments.bam --genome genome.fa\n"
-"Train a methylation model\n"
+"Test CpG sites for methylation\n"
 "\n"
 "  -v, --verbose                        display verbose output\n"
 "      --version                        display version\n"
@@ -106,6 +107,7 @@ namespace opt
     static std::string genome_file;
     static std::string models_fofn;
     static std::string region;
+    static std::string cpg_methylation_model_type = "reftrained";
     static int progress = 0;
     static int num_threads = 1;
     static int batch_size = 128;
@@ -129,35 +131,20 @@ static const struct option longopts[] = {
     { NULL, 0, NULL, 0 }
 };
 
-// Expand the event indices outwards
-void bump(HMMInputData& data, int amount)
-{
-    if(data.event_start_idx < data.event_stop_idx) {
-        data.event_start_idx -= amount;
-        data.event_stop_idx += amount;
-    } else {
-        data.event_start_idx += amount;
-        data.event_stop_idx -= amount;
-    }
-}
-// Realign the read in event space
-void test_read(const ModelMap& model_map,
-               const Fast5Map& name_map, 
-               const faidx_t* fai, 
-               const bam_hdr_t* hdr, 
-               const bam1_t* record, 
-               size_t read_idx,
-               const OutputHandles& handles)
+// Test CpG sites in this read for methylation
+void calculate_methylation_for_read(const Fast5Map& name_map,
+                                    const faidx_t* fai,
+                                    const bam_hdr_t* hdr,
+                                    const bam1_t* record,
+                                    size_t read_idx,
+                                    const OutputHandles& handles)
 {
     // Load a squiggle read for the mapped read
     std::string read_name = bam_get_qname(record);
     std::string fast5_path = name_map.get_path(read_name);
-
-    // load read
     SquiggleRead sr(read_name, fast5_path);
-    double read_score = 0.0f;
-    size_t num_sites_tested = 0;
 
+    // An output map from reference positions to scored CpG sites
     std::map<int, ScoredSite> site_score_map;
 
     for(size_t strand_idx = 0; strand_idx < NUM_STRANDS; ++strand_idx) {
@@ -166,22 +153,13 @@ void test_read(const ModelMap& model_map,
         std::vector<int> site_ends;
         std::vector<int> site_count;
 
-        double strand_score = 0.0f;
+        // replace the baked-in pore model with the methylation model
+        // (including unmethylated kmers) for this strand
+        sr.replace_model(strand_idx, opt::cpg_methylation_model_type);
 
-        // replace model 
-        std::string curr_model = sr.pore_model[strand_idx].name;
+        size_t k = sr.pore_model[strand_idx].k;
 
-        std::string methyl_model = curr_model + ".methyltrain";
-        auto model_iter = model_map.find(methyl_model);
-
-        if(model_iter != model_map.end()) {
-            sr.pore_model[strand_idx].update_states( model_iter->second );
-        } else {
-            fprintf(stderr, "Error, methylated model %s not found\n", methyl_model.c_str());
-            exit(EXIT_FAILURE);
-        }
-
-        // Align to the new model
+        // Align in event space using the new model
         EventAlignmentParameters params;
         params.sr = &sr;
         params.fai = fai;
@@ -196,8 +174,6 @@ void test_read(const ModelMap& model_map,
             continue;
         std::string contig = alignment_output.front().ref_name.c_str();
         
-        //emit_event_alignment_tsv(stdout, sr, params, alignment_output);
-        
         // Convert the EventAlignment to a map between reference positions and events
         std::vector<AlignedPair> event_aligned_pairs;
         for(size_t i = 0; i < alignment_output.size(); ++i) {
@@ -210,13 +186,13 @@ void test_read(const ModelMap& model_map,
         int ref_start_pos = event_aligned_pairs.front().ref_pos;
         int ref_end_pos = event_aligned_pairs.back().ref_pos;
 
+        // Extract the reference sequence for this region
         int fetched_len = 0;
         assert(ref_end_pos >= ref_start_pos);
-
-        // Extract the reference sequence for this region
         std::string ref_seq = get_reference_region_ts(params.fai, contig.c_str(), ref_start_pos, 
                                                   ref_end_pos, &fetched_len);
         
+        // Remove non-ACGT bases from this reference segment
         ref_seq = gDNAAlphabet.disambiguate(ref_seq);
 
         // Scan the sequence for CpGs
@@ -228,11 +204,12 @@ void test_read(const ModelMap& model_map,
             }
         }
         
-        // Batch the CpGs together
+        // Batch the CpGs together into groups that are separated by some minimum distance
         int min_separation = 10;
         size_t curr_idx = 0;
         while(curr_idx < cpg_sites.size()) {
-
+            
+            // Find the endpoint of this group of sites
             size_t end_idx = curr_idx + 1;
             while(end_idx < cpg_sites.size()) {
                 if(cpg_sites[end_idx] - cpg_sites[end_idx - 1] > min_separation)
@@ -240,6 +217,7 @@ void test_read(const ModelMap& model_map,
                 end_idx += 1; 
             }
 
+            // the coordinates on the reference substring for this group of sites
             int sub_start_pos = cpg_sites[curr_idx] - min_separation;
             int sub_end_pos = cpg_sites[end_idx - 1] + min_separation;
 
@@ -248,16 +226,19 @@ void test_read(const ModelMap& model_map,
                 std::string subseq = ref_seq.substr(sub_start_pos, sub_end_pos - sub_start_pos + 1);
                 std::string rc_subseq = mtest_alphabet->reverse_complement(subseq);
 
+                // using the reference-to-event map, look up the event indices for this segment
                 AlignedPairRefLBComp lb_comp;
                 AlignedPairConstIter start_iter = std::lower_bound(event_aligned_pairs.begin(), event_aligned_pairs.end(),
                                                                    sub_start_pos + ref_start_pos, lb_comp);
 
                 AlignedPairConstIter stop_iter = std::lower_bound(event_aligned_pairs.begin(), event_aligned_pairs.end(),
-                                                                   sub_end_pos + ref_start_pos, lb_comp);
-
-                if(start_iter != event_aligned_pairs.end() && stop_iter != event_aligned_pairs.end()) {
-
-                    std::string site_string = ref_seq.substr(cpg_sites[curr_idx] - 3, 5);
+                                                                  sub_end_pos + ref_start_pos, lb_comp);
+                
+                // Only process this region if the the read is aligned within the boundaries
+                // and the span between the start/end is not unusually short
+                if(start_iter != event_aligned_pairs.end() && stop_iter != event_aligned_pairs.end() &&
+                    abs(start_iter->read_pos - stop_iter->read_pos) > 10) 
+                {
                     
                     uint32_t hmm_flags = HAF_ALLOW_PRE_CLIP | HAF_ALLOW_POST_CLIP;
 
@@ -269,23 +250,21 @@ void test_read(const ModelMap& model_map,
                     data.rc = alignment_output.front().rc;
                     data.event_start_idx = start_iter->read_pos;
                     data.event_stop_idx = stop_iter->read_pos;
-                    data.event_stride = data.event_start_idx < data.event_stop_idx ? 1 : -1;
-                    
+                    data.event_stride = data.event_start_idx <= data.event_stop_idx ? 1 : -1;
+                 
+                    // Calculate the likelihood of the unmethylated sequence
                     HMMInputSequence unmethylated(subseq, rc_subseq, mtest_alphabet);
                     double unmethylated_score = profile_hmm_score(unmethylated, data, hmm_flags);
 
-                    // Methylate the CpGs in the sequence and score again
-                    std::string mcpg_subseq = gMCpGAlphabet.methylate(subseq);
-                    std::string rc_mcpg_subseq = gMCpGAlphabet.reverse_complement(mcpg_subseq);
-                    
-                    //printf("m_subs: %s\n", mcpg_subseq.c_str());
-                    //printf("m_rc_s: %s\n", rc_mcpg_subseq.c_str());
+                    // Methylate all CpGs in the sequence and score again
+                    std::string mcpg_subseq = mtest_alphabet->methylate(subseq);
+                    std::string rc_mcpg_subseq = mtest_alphabet->reverse_complement(mcpg_subseq);
                     
+                    // Calculate the likelihood of the methylated sequence
                     HMMInputSequence methylated(mcpg_subseq, rc_mcpg_subseq, mtest_alphabet);
                     double methylated_score = profile_hmm_score(methylated, data, hmm_flags);
-                    double diff = methylated_score - unmethylated_score;
 
-                    ScoredSite ss;
+                    // Aggregate score
                     int start_position = cpg_sites[curr_idx] + ref_start_pos;
                     auto iter = site_score_map.find(start_position);
                     if(iter == site_score_map.end()) {
@@ -295,23 +274,20 @@ void test_read(const ModelMap& model_map,
                         ss.start_position = start_position;
                         ss.end_position = cpg_sites[end_idx - 1] + ref_start_pos;
                         ss.n_cpg = end_idx - curr_idx;
-                        ss.sequence = site_string;
+
+                        // extract the CpG site(s) with a k-mers worth of surrounding context
+                        size_t site_output_start = cpg_sites[curr_idx] - k + 1;
+                        size_t site_output_end =  cpg_sites[end_idx - 1] + k;
+                        ss.sequence = ref_seq.substr(site_output_start, site_output_end - site_output_start);
+                    
+                        // insert into the map    
                         iter = site_score_map.insert(std::make_pair(start_position, ss)).first;
                     }
                     
+                    // set strand-specific score
+                    // upon output below the strand scores will be summed
                     iter->second.ll_unmethylated[strand_idx] = unmethylated_score;
                     iter->second.ll_methylated[strand_idx] = methylated_score;
-
-                    /*
-                    // Debug alignments
-                    printf("Forward unmethylated: %.2lf\n", unmethylated_score);
-                    printf("Forward methylated: %.2lf\n", methylated_score);
-                    std::vector<HMMAlignmentState> um_align = profile_hmm_align(unmethylated, data, hmm_flags);
-                    print_alignment("unmethylated", start_position, 0, unmethylated, data, um_align);
-                    
-                    std::vector<HMMAlignmentState> m_align = profile_hmm_align(methylated, data, hmm_flags);
-                    print_alignment("methylated", start_position, 0, methylated, data, m_align);
-                    */
                 }
             }
 
@@ -321,10 +297,12 @@ void test_read(const ModelMap& model_map,
     
     #pragma omp critical(methyltest_write)
     {
+        // these variables are sums over all sites within a read
         double ll_ratio_sum_strand[2] = { 0.0f, 0.0f };
         double ll_ratio_sum_both = 0;
         size_t num_positive = 0;
 
+        // write all sites for this read
         for(auto iter = site_score_map.begin(); iter != site_score_map.end(); ++iter) {
 
             const ScoredSite& ss = iter->second;
@@ -336,15 +314,18 @@ void test_read(const ModelMap& model_map,
             num_positive += diff > 0;
 
             fprintf(handles.site_writer, "%s\t%d\t%d\t", ss.chromosome.c_str(), ss.start_position, ss.end_position);
-            fprintf(handles.site_writer, "LL_METH=%.2lf;LL_UNMETH=%.2lf;LL_RATIO=%.2lf;", sum_ll_m, sum_ll_u, diff);
-            fprintf(handles.site_writer, "N_CPG=%d;SEQUENCE=%s\n", ss.n_cpg, ss.sequence.c_str());
+            fprintf(handles.site_writer, "ReadIdx=%zu;", read_idx);
+            fprintf(handles.site_writer, "LogLikMeth=%.2lf;LogLikUnmeth=%.2lf;LogLikRatio=%.2lf;", sum_ll_m, sum_ll_u, diff);
+            fprintf(handles.site_writer, "LogLikMethByStrand=%.2lf,%.2lf;", ss.ll_methylated[0], ss.ll_methylated[1]);
+            fprintf(handles.site_writer, "LogLikUnmethByStrand=%.2lf,%.2lf;", ss.ll_unmethylated[0], ss.ll_unmethylated[1]);
+            fprintf(handles.site_writer, "NumCpGs=%d;Sequence=%s\n", ss.n_cpg, ss.sequence.c_str());
 
             ll_ratio_sum_strand[0] += ss.ll_methylated[0] - ss.ll_unmethylated[0];
             ll_ratio_sum_strand[1] += ss.ll_methylated[1] - ss.ll_unmethylated[1];
             ll_ratio_sum_both += diff;
         }
         std::string complement_model = sr.pore_model[C_IDX].name;
-        fprintf(handles.read_writer, "%s\t%.2lf\t%zu\t%s\tNUM_POSITIVE=%zu\n", fast5_path.c_str(), ll_ratio_sum_both, site_score_map.size(), complement_model.c_str(), num_positive);
+        fprintf(handles.read_writer, "%s\t%.2lf\t%zu\t%s\tNumPositive=%zu\n", fast5_path.c_str(), ll_ratio_sum_both, site_score_map.size(), complement_model.c_str(), num_positive);
     
         for(size_t si = 0; si < NUM_STRANDS; ++si) {
             std::string model = sr.pore_model[si].name;
@@ -394,7 +375,7 @@ void parse_methyltest_options(int argc, char** argv)
         std::cerr << SUBPROGRAM ": a --reads file must be provided\n";
         die = true;
     }
-    
+
     if(opt::genome_file.empty()) {
         std::cerr << SUBPROGRAM ": a --genome file must be provided\n";
         die = true;
@@ -404,13 +385,16 @@ void parse_methyltest_options(int argc, char** argv)
         std::cerr << SUBPROGRAM ": a --bam file must be provided\n";
         die = true;
     }
-    
+
     if(opt::models_fofn.empty()) {
-        std::cerr << SUBPROGRAM ": a --models-fofn file must be provided\n";
+        std::cerr << SUBPROGRAM ": a --models file must be provided\n";
         die = true;
+    } else {
+        // initialize the model set from the fofn
+        PoreModelSet::initialize(opt::models_fofn);
     }
 
-    if (die) 
+    if (die)
     {
         std::cout << "\n" << METHYLTEST_USAGE_MESSAGE;
         exit(EXIT_FAILURE);
@@ -423,8 +407,7 @@ int methyltest_main(int argc, char** argv)
     omp_set_num_threads(opt::num_threads);
 
     Fast5Map name_map(opt::reads_file);
-    ModelMap models = read_models_fofn(opt::models_fofn);
-    
+
     // Open the BAM and iterate over reads
 
     // load bam file
@@ -438,7 +421,7 @@ int methyltest_main(int argc, char** argv)
 
     // read the bam header
     bam_hdr_t* hdr = sam_hdr_read(bam_fh);
-    
+
     // load reference fai file
     faidx_t *fai = fai_load(opt::genome_file.c_str());
 
@@ -492,20 +475,20 @@ int methyltest_main(int argc, char** argv)
 
     do {
         assert(num_records_buffered < records.size());
-        
+
         // read a record into the next slot in the buffer
         result = sam_itr_next(bam_fh, itr, records[num_records_buffered]);
         num_records_buffered += result >= 0;
 
         // realign if we've hit the max buffer size or reached the end of file
         if(num_records_buffered == records.size() || result < 0) {
-            
+
             #pragma omp parallel for
             for(size_t i = 0; i < num_records_buffered; ++i) {
                 bam1_t* record = records[i];
                 size_t read_idx = num_reads_processed + i;
                 if( (record->core.flag & BAM_FUNMAP) == 0) {
-                    test_read(models, name_map, fai, hdr, record, read_idx, handles);
+                    calculate_methylation_for_read(name_map, fai, hdr, record, read_idx, handles);
                 }
             }
 
@@ -514,7 +497,7 @@ int methyltest_main(int argc, char** argv)
 
         }
     } while(result >= 0);
-    
+
     assert(num_records_buffered == 0);
     progress.end();
 
